import asyncio
import datetime
import pickle
import re
import traceback
import sys
from pathlib import Path


import aiofiles
import json

import discord
from discord.ext import pages

from models.deepl_model import TranslationModel
from models.embed_statics_model import EmbedStatics
from models.openai_model import Override
from services.environment_service import EnvService
from services.message_queue_service import Message
from services.moderations_service import Moderation
from models.user_model import Thread, EmbeddedConversationItem, Instruction
from collections import defaultdict
from sqlitedict import SqliteDict

from services.pickle_service import Pickler
from services.sharegpt_service import ShareGPTService
from services.text_service import SetupModal, TextService

original_message = {}
ALLOWED_GUILDS = EnvService.get_allowed_guilds()
if sys.platform == "win32":
    separator = "\\"
else:
    separator = "/"

#
# Get the user key service if it is enabled.
#
USER_INPUT_API_KEYS = EnvService.get_user_input_api_keys()
USER_KEY_DB = EnvService.get_api_db()
CHAT_BYPASS_ROLES = EnvService.get_bypass_roles()
PRE_MODERATE = EnvService.get_premoderate()
FORCE_ENGLISH = EnvService.get_force_english()
BOT_TAGGABLE = EnvService.get_bot_is_taggable()
CHANNEL_CHAT_ROLES = EnvService.get_channel_chat_roles()
BOT_TAGGABLE_ROLES = EnvService.get_gpt_roles()
CHANNEL_INSTRUCTION_ROLES = EnvService.get_channel_instruction_roles()

#
# Obtain the Moderation table and the General table, these are two SQLite tables that contain
# information about the server that are used for persistence and to auto-restart the moderation service.
#
MOD_DB = None
GENERAL_DB = None
try:
    print("Attempting to retrieve the General and Moderations DB")
    MOD_DB = SqliteDict(
        EnvService.find_shared_file("main_db.sqlite"),
        tablename="moderations",
        autocommit=True,
    )
    GENERAL_DB = SqliteDict(
        EnvService.find_shared_file("main_db.sqlite"),
        tablename="general",
        autocommit=True,
    )
    print("Retrieved the General and Moderations DB")
except Exception as e:
    print("Failed to retrieve the General and Moderations DB. The bot is terminating.")
    raise e

BOT_NAME = EnvService.get_custom_bot_name()


class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
    def __init__(
        self,
        bot,
        usage_service,
        model,
        message_queue,
        deletion_queue,
        DEBUG_GUILD,
        DEBUG_CHANNEL,
        data_path: Path,
        pinecone_service,
        pickle_queue,
    ):
        super().__init__()
        self.GLOBAL_COOLDOWN_TIME = 0.25

        # Environment
        self.data_path = data_path
        self.debug_channel = None

        # Services and models
        self.bot = bot
        self.usage_service = usage_service
        self.model = model
        self.translation_model = TranslationModel()
        self.deletion_queue = deletion_queue

        # Data specific to all text based GPT interactions
        self.users_to_interactions = defaultdict(list)
        self.redo_users = {}

        # Pickle queue
        self.pickle_queue = pickle_queue

        # Conversations-specific data
        self.END_PROMPTS = [
            "end",
            "end conversation",
            "end the conversation",
            "that's all",
            "that'll be all",
        ]
        self.awaiting_responses = []
        self.awaiting_thread_responses = []
        self.conversation_threads = {}
        self.full_conversation_history = defaultdict(list)
        self.instructions = defaultdict(list)
        self.summarize = self.model.summarize_conversations

        # Pinecone data
        self.pinecone_service = pinecone_service

        # Sharing service
        self.sharegpt_service = ShareGPTService()

        try:
            conversation_file_path = EnvService.find_shared_file(
                "conversation_starter_pretext.txt"
            )
            # Attempt to read a conversation starter text string from the file.
            with conversation_file_path.open("r") as f:
                self.CONVERSATION_STARTER_TEXT = f.read()
                print(
                    f"Conversation starter text loaded from {conversation_file_path}."
                )
            assert self.CONVERSATION_STARTER_TEXT is not None

            language_detect_file_path = EnvService.find_shared_file(
                "language_detection_pretext.txt"
            )
            # Attempt to read a conversation starter text string from the file.
            with language_detect_file_path.open("r") as f:
                self.LANGUAGE_DETECT_STARTER_TEXT = f.read()
                print(
                    f"Language detection starter text loaded from {language_detect_file_path}."
                )
            assert self.LANGUAGE_DETECT_STARTER_TEXT is not None

            conversation_file_path_minimal = EnvService.find_shared_file(
                "conversation_starter_pretext_minimal.txt"
            )
            with conversation_file_path_minimal.open("r") as f:
                self.CONVERSATION_STARTER_TEXT_MINIMAL = f.read()
                print(
                    f"Conversation starter text loaded from {conversation_file_path_minimal}."
                )
            assert self.CONVERSATION_STARTER_TEXT_MINIMAL is not None

        except Exception:
            self.CONVERSATION_STARTER_TEXT = self.CONVERSATION_STARTER_TEXT_MINIMAL = (
                "You are an artificial intelligence that is able to do anything, and answer any question,"
                "I want you to be my personal assistant and help me with some tasks. "
                "and I want you to make well-informed decisions using the data that you have been trained on, "
                "and be sure to be mindful of the previous conversation history and be consistent with your answers."
            )

        self.DEBUG_GUILD = DEBUG_GUILD
        self.DEBUG_CHANNEL = DEBUG_CHANNEL
        print(
            f"The debug channel and guild IDs are {self.DEBUG_GUILD} and {self.DEBUG_CHANNEL}"
        )
        self.TEXT_CUTOFF = 1900
        self.EMBED_CUTOFF = 3900
        self.message_queue = message_queue
        self.conversation_thread_owners = defaultdict(list)

    async def load_file(self, file, ctx):
        """Take filepath, return content or respond if not found"""
        try:
            async with aiofiles.open(file, "r") as f:
                return await f.read()
        except Exception as e:
            traceback.print_exc()
            await ctx.respond(
                "Error loading file. Please check that it is correctly placed in the bot's root file directory."
            )
            raise e

    @discord.Cog.listener()
    async def on_member_join(self, member):
        """When members join send welcome message if enabled"""
        if self.model.welcome_message_enabled:
            query = f"Please generate a welcome message for {member.name} who has just joined the server."

            try:
                welcome_message_response = await self.model.send_request(
                    query,
                    tokens=self.usage_service.count_tokens(query),
                    is_chatgpt_request=True
                    if "turbo" in str(self.model.model)
                    else False,
                )
                welcome_message = str(welcome_message_response["choices"][0]["text"])
            except Exception:
                welcome_message = None

            if not welcome_message:
                welcome_message = EnvService.get_welcome_message()
            welcome_embed = discord.Embed(
                title=f"Welcome, {member.name}!", description=welcome_message
            )

            welcome_embed.add_field(
                name="Just so you know...",
                value="> My commands are invoked with a forward slash (/)\n> Use /help to see my help message(s).",
            )
            await member.send(content=None, embed=welcome_embed)

    @discord.Cog.listener()
    async def on_ready(self):
        """When ready to recieve data set debug channel and sync commands"""
        self.debug_channel = self.bot.get_guild(self.DEBUG_GUILD).get_channel(
            self.DEBUG_CHANNEL
        )
        print("The debug channel was acquired")

        print("Attempting to load from pickles")
        # Try to load self.full_conversation_history, self.conversation_threads, and self.conversation_thread_owners from the `pickles` folder
        try:
            with open(
                EnvService.save_path() / "pickles" / "full_conversation_history.pickle",
                "rb",
            ) as f:
                self.full_conversation_history = pickle.load(f)
                print("Loaded full_conversation_history")

            with open(
                EnvService.save_path() / "pickles" / "conversation_threads.pickle", "rb"
            ) as f:
                self.conversation_threads = pickle.load(f)
                print("Loaded conversation_threads")

            with open(
                EnvService.save_path()
                / "pickles"
                / "conversation_thread_owners.pickle",
                "rb",
            ) as f:
                self.conversation_thread_owners = pickle.load(f)
                print("Loaded conversation_thread_owners")

            with open(
                EnvService.save_path() / "pickles" / "instructions.pickle",
                "rb",
            ) as f:
                self.instructions = pickle.load(f)
                print("Loaded instructions")

            # Fail if all three weren't loaded
            assert self.full_conversation_history is not {}
            assert self.conversation_threads is not {}
            assert self.conversation_thread_owners is not defaultdict(list)

        except Exception:
            print("Failed to load from pickles")
            self.full_conversation_history = defaultdict(list)
            self.conversation_threads = {}
            self.conversation_thread_owners = defaultdict(list)
            traceback.print_exc()

        print("Syncing commands...")

        await self.bot.sync_commands(
            commands=None,
            method="individual",
            force=True,
            guild_ids=ALLOWED_GUILDS,
            register_guild_commands=True,
            check_guilds=[],
            delete_existing=True,
        )
        print("Commands synced")

        # Start an inline async loop that runs every 10 seconds to save the conversation history to a pickle file
        print("Starting pickle loop")
        while True:
            await asyncio.sleep(15)
            await self.pickle_queue.put(
                Pickler(
                    self.full_conversation_history,
                    self.conversation_threads,
                    self.conversation_thread_owners,
                    self.instructions,
                )
            )

    def check_conversing(self, channel_id, message_content):
        '''given channel id and a message, return true if it's a conversation thread, false if not, or if the message starts with "~"'''
        cond1 = channel_id in self.conversation_threads
        # If the trimmed message starts with a Tilde, then we want to not contribute this to the conversation
        try:
            cond2 = not message_content.strip().startswith("~")
        except Exception as e:
            print(e)
            cond2 = False

        return (cond1) and cond2

    async def end_conversation(
        self, ctx, opener_user_id=None, conversation_limit=False
    ):
        """end the thread of the user interacting with the bot, if the conversation has reached the limit close it for the owner"""
        normalized_user_id = opener_user_id if opener_user_id else ctx.author.id
        # Check if the channel is an instance of a thread
        thread = False
        if isinstance(ctx.channel, discord.Thread):
            thread = True

        if (
            conversation_limit
        ):  # if we reach the conversation limit we want to close from the channel it was maxed out in
            channel_id = ctx.channel.id
        else:
            try:
                channel_ids = self.conversation_thread_owners[normalized_user_id]
                if ctx.channel.id not in channel_ids:
                    await ctx.reply(
                        "This is not a conversation thread that you own!",
                        delete_after=5,
                    )
                    return

                if normalized_user_id in self.awaiting_responses:
                    await ctx.reply(
                        embed=discord.Embed(
                            title=f"Please wait for a response before ending the conversation.",
                            color=0x808080,
                        )
                    )
                    return

            except Exception:
                traceback.print_exc()
                await ctx.delete(delay=5)
                await ctx.reply(
                    "Only the conversation starter can end this.", delete_after=5
                )
                return

        # TODO Possible bug here, if both users have a conversation active and one user tries to end the other, it may
        # allow them to click the end button on the other person's thread and it will end their own convo.
        self.conversation_threads.pop(ctx.channel.id)

        if isinstance(
            ctx, discord.ApplicationContext
        ):  # When the conversation is ended from the slash command
            await ctx.respond(
                "You have ended the conversation with GPT. Start a conversation with /gpt converse",
                ephemeral=True,
                delete_after=10,
            )
        elif isinstance(
            ctx, discord.Interaction
        ):  # When the user ends the conversation from the button
            await ctx.response.send_message(
                "You have ended the conversation with GPT. Start a conversation with /gpt converse",
                ephemeral=True,
                delete_after=10,
            )
        else:  # The case for when the user types "end" in the channel
            await ctx.reply(
                "You have ended the conversation with GPT. Start a conversation with /gpt converse",
                delete_after=10,
            )

        await ctx.channel.send(
            embed=EmbedStatics.generate_end_embed(),
            view=ShareView(self, ctx.channel.id) if thread else None,
        )

        # Close all conversation threads for the user
        # If at conversation limit then fetch the owner and close the thread for them
        if conversation_limit:
            try:
                owner_id = [
                    owner
                    for owner, threads in self.conversation_thread_owners.items()
                    if channel_id in threads
                ][0]
                self.conversation_thread_owners[owner_id].remove(ctx.channel.id)
                # Attempt to close and lock the thread.
                if thread:
                    try:
                        thread = await self.bot.fetch_channel(channel_id)
                        await thread.edit(name="Closed-GPT")
                        await thread.edit(archived=True)
                    except Exception:
                        traceback.print_exc()
            except Exception:
                traceback.print_exc()
        else:
            if normalized_user_id in self.conversation_thread_owners:
                thread_id = ctx.channel.id
                self.conversation_thread_owners[normalized_user_id].remove(
                    ctx.channel.id
                )

                # Attempt to close and lock the thread.
                if thread:
                    try:
                        thread = await self.bot.fetch_channel(thread_id)
                        await thread.edit(name="Closed-GPT")
                        await thread.edit(archived=True)
                    except Exception:
                        traceback.print_exc()

    async def send_settings_text(self, ctx):
        """compose and return the settings menu to the interacting user"""
        embed = discord.Embed(
            title="GPT3Bot Settings",
            description="The current settings of the model",
            color=0x00FF00,
        )
        # Create a two-column embed to display the settings, use \u200b to create a blank space
        embed.add_field(
            name="Setting",
            value="\n".join(
                [
                    key
                    for key in self.model.__dict__.keys()
                    if key not in self.model._hidden_attributes
                ]
            ),
            inline=True,
        )
        embed.add_field(
            name="Value",
            value="\n".join(
                [
                    str(value)
                    for key, value in self.model.__dict__.items()
                    if key not in self.model._hidden_attributes
                ]
            ),
            inline=True,
        )
        await ctx.respond(embed=embed, ephemeral=True)

    async def process_settings(self, ctx, parameter, value):
        """Given a parameter and value set the corresponding parameter in storage to the value"""

        # Check if the parameter is a valid parameter
        if hasattr(self.model, parameter):
            # Check if the value is a valid value
            try:
                # Set the parameter to the value
                setattr(self.model, parameter, value)
                await ctx.respond(
                    "Successfully set the parameter " + parameter + " to " + value
                )

                if parameter == "mode":
                    await ctx.send_followup(
                        "The mode has been set to "
                        + value
                        + ". This has changed the temperature top_p to the mode defaults of "
                        + str(self.model.temp)
                        + " and "
                        + str(self.model.top_p)
                    )
            except ValueError as e:
                await ctx.respond(e)
        else:
            await ctx.respond("The parameter is not a valid parameter")

    def generate_debug_message(self, prompt, response):
        """create a debug message with a prompt and a response field"""
        debug_message = "----------------------------------------------------------------------------------\n"
        debug_message += "Prompt:\n```\n" + prompt + "\n```\n"
        debug_message += "Response:\n```\n" + json.dumps(response, indent=4) + "\n```\n"
        return debug_message

    async def paginate_and_send(self, response_text, ctx):
        """paginate a response to a text cutoff length and send it in chunks"""
        from_context = isinstance(ctx, discord.ApplicationContext)

        response_text = [
            response_text[i : i + self.TEXT_CUTOFF]
            for i in range(0, len(response_text), self.TEXT_CUTOFF)
        ]
        # Send each chunk as a message
        first = False
        for chunk in response_text:
            if not first:
                if from_context:
                    await ctx.send_followup(chunk)
                else:
                    await ctx.reply(chunk)
                first = True
            else:
                if from_context:
                    response_message = await ctx.send_followup(chunk)
                else:
                    response_message = await ctx.channel.send(chunk)
        return response_message

    async def paginate_embed(self, response_text):
        """Given a response text make embed pages and return a list of the pages."""

        response_text = [
            response_text[i : i + self.EMBED_CUTOFF]
            for i in range(0, len(response_text), self.EMBED_CUTOFF)
        ]
        pages = []
        first = False
        # Send each chunk as a message
        for count, chunk in enumerate(response_text, start=1):
            if not first:
                page = discord.Embed(
                    title=f"Page {count}",
                    description=chunk,
                )
                first = True
            else:
                page = discord.Embed(
                    title=f"Page {count}",
                    description=chunk,
                )
            pages.append(page)

        return pages

    async def queue_debug_message(self, debug_message, debug_channel):
        """Put a message into the debug queue"""
        await self.message_queue.put(Message(debug_message, debug_channel))

    async def queue_debug_chunks(self, debug_message, debug_channel):
        """Put a message as chunks into the debug queue"""
        debug_message_chunks = [
            debug_message[i : i + self.TEXT_CUTOFF]
            for i in range(0, len(debug_message), self.TEXT_CUTOFF)
        ]

        backticks_encountered = 0

        for i, chunk in enumerate(debug_message_chunks):
            # Count the number of backticks in the chunk
            backticks_encountered += chunk.count("```")

            # If it's the first chunk, append a "\n```\n" to the end
            if i == 0:
                chunk += "\n```\n"

            # If it's an interior chunk, append a "```\n" to the end, and a "\n```\n" to the beginning
            elif i < len(debug_message_chunks) - 1:
                chunk = "\n```\n" + chunk + "```\n"

            # If it's the last chunk, append a "```\n" to the beginning
            else:
                chunk = "```\n" + chunk

            await self.message_queue.put(Message(chunk, debug_channel))

    async def send_debug_message(self, debug_message, debug_channel):
        """process a debug message and put directly into queue or chunk it"""
        # Send the debug message
        try:
            if len(debug_message) > self.TEXT_CUTOFF:
                await self.queue_debug_chunks(debug_message, debug_channel)
            else:
                await self.queue_debug_message(debug_message, debug_channel)
        except Exception as e:
            traceback.print_exc()
            await self.message_queue.put(
                Message("Error sending debug message: " + str(e), debug_channel)
            )

    async def check_conversation_limit(self, message):
        """Check if a conversation has reached the set limit and end it if it has"""
        # After each response, check if the user has reached the conversation limit in terms of messages or time.
        if message.channel.id in self.conversation_threads:
            # If the user has reached the max conversation length, end the conversation
            if (
                self.conversation_threads[message.channel.id].count
                >= self.model.max_conversation_length
            ):
                await message.reply(
                    "You have reached the maximum conversation length. You have ended the conversation with GPT, and it has ended."
                )
                await self.end_conversation(message, conversation_limit=True)
                return True
        return False

    async def summarize_conversation(self, message, prompt):
        """Takes a conversation history filled prompt and summarizes it to then start a new history with it as the base"""
        response = await self.model.send_summary_request(prompt)
        summarized_text = response["choices"][0]["text"]

        new_conversation_history = []
        new_conversation_history.append(
            EmbeddedConversationItem(self.CONVERSATION_STARTER_TEXT, 0)
        )
        new_conversation_history.append(
            EmbeddedConversationItem(
                f"\nThis conversation has some context from earlier, which has been summarized as follows: {summarized_text} \nContinue the conversation, paying very close attention to things <username> told you, such as their name, and personal details.",
                0,
            )
        )

        # Get the last entry from the thread's conversation history
        new_conversation_history.append(
            EmbeddedConversationItem(
                self.conversation_threads[message.channel.id].history[-1].text + "\n", 0
            )
        )
        self.conversation_threads[message.channel.id].history = new_conversation_history

    # A listener for message edits to redo prompts if they are edited
    @discord.Cog.listener()
    async def on_message_edit(self, before, after):
        """When a message is edited run moderation if enabled, and process if it a prompt that should be redone"""

        if after.author.id == self.bot.user.id:
            return

        # Moderation
        if not isinstance(after.channel, discord.DMChannel):
            if (
                after.guild.id in Moderation.moderation_queues
                and Moderation.moderation_queues[after.guild.id] is not None
            ):
                # Create a timestamp that is 0.25 seconds from now
                timestamp = (
                    datetime.datetime.now() + datetime.timedelta(seconds=0.25)
                ).timestamp()
                await Moderation.moderation_queues[after.guild.id].put(
                    Moderation(after, timestamp)
                )  # TODO Don't proceed if message was deleted!

        await TextService.process_conversation_edit(self, after, original_message)

    @discord.Cog.listener()
    async def on_message(self, message):
        """On a new message check if it should be moderated then process it for conversation"""
        if message.author == self.bot.user:
            return

        # Check if the message is a discord system message
        if message.type != discord.MessageType.default:
            return

        # Moderations service is done here.
        if (
            hasattr(message, "guild")
            and message.guild.id in Moderation.moderation_queues
            and Moderation.moderation_queues[message.guild.id] is not None
        ):
            # Don't moderate if there is no "roles" attribute for the author
            if not hasattr(message.author, "roles"):
                pass
            # Verify that the user is not in a role that can bypass moderation
            elif CHAT_BYPASS_ROLES is [None] or not any(
                role.name.lower() in CHAT_BYPASS_ROLES for role in message.author.roles
            ):
                # Create a timestamp that is 0.5 seconds from now
                timestamp = (
                    datetime.datetime.now() + datetime.timedelta(seconds=0.5)
                ).timestamp()
                await Moderation.moderation_queues[message.guild.id].put(
                    Moderation(message, timestamp)
                )

        # Language check
        if FORCE_ENGLISH and len(message.content.split(" ")) > 3:
            if not await Moderation.force_english_and_respond(
                message.content, self.LANGUAGE_DETECT_STARTER_TEXT, message
            ):
                await message.delete()
                return

        # Get the first file in the message if there is one
        file = message.attachments[0] if len(message.attachments) > 0 else None

        # Process the message if the user is in a conversation
        if await TextService.process_conversation_message(
            self, message, USER_INPUT_API_KEYS, USER_KEY_DB, file=file
        ):
            original_message[message.author.id] = message.id

        # If the user tagged the bot and the tag wasn't an @here or @everyone, retrieve the message
        if f"<@{self.bot.user.id}>" in message.content and not (
            "@everyone" in message.content or "@here" in message.content
        ):
            if not BOT_TAGGABLE:
                return

            # Check if any of the message author's role names are in BOT_TAGGABLE_ROLES, if not, return
            if BOT_TAGGABLE_ROLES != [None] and not any(
                role.name.lower() in BOT_TAGGABLE_ROLES for role in message.author.roles
            ):
                return

            # Remove the mention from the message
            prompt = message.content.replace(self.bot.user.mention, "")
            # If the message is empty, don't process it
            if len(prompt) < 5:
                await message.reply(
                    "This is too short of a prompt to think about. Please be more specific."
                )
                return

            await self.ask_command(
                message,
                prompt=prompt,
                from_message_context=True,
            )

    def cleanse_response(self, response_text):
        """Cleans history tokens from response"""
        response_text = response_text.replace("<yourname>:", "")
        response_text = response_text.replace("You:", "")
        response_text = response_text.replace(BOT_NAME.replace(" ", ""), "")
        response_text = response_text.replace(BOT_NAME, "")
        response_text = response_text.replace("<|endofstatement|>", "")
        return response_text

    def remove_awaiting(
        self, author_id, channel_id, from_ask_command, from_edit_command
    ):
        """Remove user from ask/edit command response wait, if not any of those then process the id to remove user from thread response wait"""
        if author_id in self.awaiting_responses:
            self.awaiting_responses.remove(author_id)
        if not from_ask_command and not from_edit_command:
            if channel_id in self.awaiting_thread_responses:
                self.awaiting_thread_responses.remove(channel_id)

    async def mention_to_username(self, ctx, message):
        """replaces discord mentions with their server nickname in text, if the user is not found keep the mention as is"""
        if not discord.utils.raw_mentions(message):
            return message
        for mention in discord.utils.raw_mentions(message):
            try:
                user = await discord.utils.get_or_fetch(ctx.guild, "member", mention)
                message = message.replace(f"<@{str(mention)}>", user.display_name)
            except Exception:
                pass
        return message

    # COMMANDS

    async def help_command(self, ctx):
        """Command handler. Generates a help message and sends it to the user"""
        await ctx.defer()
        embed = discord.Embed(
            title="GPT3Bot Help", description="The current commands", color=0xC730C7
        )
        embed.add_field(
            name="/search",
            value="AI-Assisted google search!",
            inline=False,
        )
        embed.add_field(
            name="/index",
            value="Indexing commands for document knowledge and querying",
            inline=False,
        )
        embed.add_field(
            name="/gpt ask",
            value="Ask GPT something. Be clear, long, and concise in your prompt. Don't waste tokens.",
            inline=False,
        )
        embed.add_field(
            name="/gpt edit",
            value="Use GPT to edit a piece of text given an instruction",
            inline=False,
        )
        embed.add_field(
            name="/gpt converse", value="Start a conversation with GPT", inline=False
        )
        embed.add_field(
            name="/gpt end",
            value="End a conversation with GPT. You can also type `end` in the conversation.",
            inline=False,
        )
        embed.add_field(
            name="/dalle draw <image prompt>",
            value="Use DALL-E2 to draw an image based on a text prompt",
            inline=False,
        )
        embed.add_field(
            name="/dalle optimize <image prompt>",
            value="Optimize an image prompt for use with DALL-E2, Midjourney, SD, etc.",
            inline=False,
        )
        embed.add_field(
            name="/system settings",
            value="Print the current settings of the model",
            inline=False,
        )
        embed.add_field(
            name="/system settings <model parameter> <value>",
            value="Change the parameter of the model named by <model parameter> to new value <value>",
            inline=False,
        )
        embed.add_field(
            name="/mod",
            value="The automatic moderations service",
            inline=False,
        )
        embed.add_field(
            name="/translate",
            value="Translate from one language to another",
            inline=False,
        )

        embed.add_field(name="/help", value="See this help text", inline=False)
        await ctx.respond(embed=embed, ephemeral=False)

    async def set_usage_command(
        self, ctx: discord.ApplicationContext, usage_amount: float
    ):
        """Command handler. Sets the usage file to the given value"""
        await ctx.defer()

        # Attempt to convert the input usage value into a float
        try:
            usage = float(usage_amount)
            await self.usage_service.set_usage(usage)
            await ctx.respond(f"Set the usage to {usage}")
        except Exception:
            await ctx.respond("The usage value must be a valid float.")
            return

    async def delete_all_conversation_threads_command(
        self, ctx: discord.ApplicationContext
    ):
        """Command handler. Deletes all threads made by the bot in the current guild"""
        await ctx.defer()

        for thread in ctx.guild.threads:
            thread_name = thread.name.lower()
            if "with gpt" in thread_name or "closed-gpt" in thread_name:
                try:
                    await thread.delete()
                except Exception:
                    pass
        await ctx.respond("All conversation threads in this server have been deleted.")

    async def usage_command(self, ctx):
        """Command handler. Responds with the current usage of the bot"""
        await ctx.defer()
        embed = discord.Embed(
            title="GPT3Bot Usage", description="The current usage", color=0x00FF00
        )
        # 1000 tokens costs 0.02 USD, so we can calculate the total tokens used from the price that we have stored
        embed.add_field(
            name="Total tokens used",
            value=str(int((await self.usage_service.get_usage() / 0.02)) * 1000),
            inline=False,
        )
        embed.add_field(
            name="Total price",
            value="$" + str(round(await self.usage_service.get_usage(), 2)),
            inline=False,
        )
        await ctx.respond(embed=embed)

    async def instruction_command(
        self,
        ctx: discord.ApplicationContext,
        mode: str,
        type: str,
        instruction: str,
        instruction_file: discord.Attachment,
        private: bool,
    ):
        """Command to let users set their own system prompt or add one to the channel"""

        await ctx.defer(ephemeral=private)

        if mode == "set" and not (instruction or instruction_file):
            await ctx.respond(
                "You must include either an **instruction** or an **instruction file**"
            )
            return

        # Check if any of the message author's role names are in CHANNEL_INSTRUCTION_ROLES, if not, continue as user
        if type == "channel" and mode in ["set", "clear"]:
            if CHANNEL_INSTRUCTION_ROLES != [None] and not any(
                role.name.lower() in CHANNEL_INSTRUCTION_ROLES
                for role in ctx.author.roles
            ):
                await ctx.respond(
                    "You don't have permisson to set the channel instruction. Defaulting to setting a user instruction"
                )
                type = "user"

        if instruction_file:
            bytestring = await instruction_file.read()
            file_instruction = bytestring.decode("utf-8")
        if instruction and instruction_file:
            instruction = f"{file_instruction}\n\n{instruction}"
        elif instruction_file:
            instruction = file_instruction

        # If premoderation is enabled, check
        if PRE_MODERATE:
            if await Moderation.simple_moderate_and_respond(instruction, ctx):
                return

        if type == "channel":
            set_id = ctx.channel.id
        else:
            set_id = ctx.user.id

        if mode == "set":
            self.instructions[set_id] = Instruction(set_id, instruction)
            await ctx.respond(f"The system instruction is set for **{type}**")
        elif mode == "get":
            try:
                instruction = self.instructions[set_id].prompt
                embed_pages = await self.paginate_embed(instruction)
                paginator = pages.Paginator(
                    pages=embed_pages, timeout=None, author_check=False
                )
                await paginator.respond(ctx.interaction)
            except Exception:
                await ctx.respond("There is no instruction set")
        elif mode == "clear":
            self.instructions.pop(set_id)
            await ctx.respond(f"The instruction has been removed for **{type}**")

    async def ask_command(
        self,
        ctx: discord.ApplicationContext,
        prompt: str,
        private: bool = False,
        temperature: float = None,
        top_p: float = None,
        frequency_penalty: float = None,
        presence_penalty: float = None,
        from_ask_action=None,
        from_other_action=None,
        from_message_context=None,
        prompt_file: discord.Attachment = None,
        model=None,
    ):
        """Command handler. Requests and returns a generation with no extras to the completion endpoint

        Args:
            ctx (discord.ApplicationContext): Command interaction
            prompt (str): A prompt to use for generation
            temperature (float): Sets the temperature override
            top_p (float): Sets the top p override
            frequency_penalty (float): Sets the frequency penalty override
            presence_penalty (float): Sets the presence penalty override
            from_action (bool, optional): Enables ephemeral. Defaults to None.
        """
        is_context = isinstance(ctx, discord.ApplicationContext)

        user = ctx.user if is_context else ctx.author

        if not (prompt or prompt_file):
            await ctx.respond(
                "You must include either a **prompt** or a **prompt file**"
            )
            return

        if prompt_file:
            bytestring = await prompt_file.read()
            file_prompt = bytestring.decode("utf-8")
        if prompt and prompt_file:
            prompt = f"{file_prompt}\n\n{prompt}"
        elif prompt_file:
            prompt = file_prompt

        prompt = await self.mention_to_username(ctx, prompt.strip())

        if len(prompt) < self.model.prompt_min_length:
            alias = ctx.respond if is_context else ctx.send
            await alias(
                f"Prompt must be greater than {self.model.prompt_min_length} characters, it is currently: {len(prompt)} characters"
            )
            return

        user_api_key = None
        if USER_INPUT_API_KEYS:
            user_api_key = await TextService.get_user_api_key(user.id, ctx, USER_KEY_DB)
            if not user_api_key:
                return

        await ctx.defer(ephemeral=private) if is_context else None

        # If premoderation is enabled, check
        if PRE_MODERATE:
            if await Moderation.simple_moderate_and_respond(prompt, ctx):
                return

        overrides = Override(temperature, top_p, frequency_penalty, presence_penalty)

        await TextService.encapsulated_send(
            self,
            user.id,
            prompt,
            ctx,
            overrides=overrides,
            from_ask_command=True,
            custom_api_key=user_api_key,
            from_ask_action=from_ask_action,
            from_other_action=from_other_action,
            from_message_context=from_message_context,
            model=model,
        )

    async def edit_command(
        self,
        ctx: discord.ApplicationContext,
        instruction: str,
        text: str,
        private: bool,
        temperature: float,
        top_p: float,
    ):
        """Command handler. Requests and returns a generation with no extras to the edit endpoint

        Args:
            ctx (discord.ApplicationContext): Command interaction
            instruction (str): The modification instructions
            text (str): The text that should be modified
            temperature (float): Sets the temperature override
            top_p (float): Sets the top p override
        """
        user = ctx.user

        text = await self.mention_to_username(ctx, text.strip())
        instruction = await self.mention_to_username(ctx, instruction.strip())

        # Validate that  all the parameters are in a good state before we send the request
        if len(instruction) < self.model.prompt_min_length:
            await ctx.respond(
                f"Instruction must be at least {self.model.prompt_min_length} characters long"
            )
            return

        user_api_key = None
        if USER_INPUT_API_KEYS:
            user_api_key = await TextService.get_user_api_key(user.id, ctx, USER_KEY_DB)
            if not user_api_key:
                return

        await ctx.defer(ephemeral=private)

        if PRE_MODERATE:
            if await Moderation.simple_moderate_and_respond(instruction + text, ctx):
                return

        overrides = Override(temperature, top_p, 0, 0)

        await TextService.encapsulated_send(
            self,
            user.id,
            prompt=text,
            ctx=ctx,
            overrides=overrides,
            instruction=instruction,
            from_edit_command=True,
            custom_api_key=user_api_key,
        )

    async def private_test_command(self, ctx: discord.ApplicationContext):
        """Command handler. Creates a private thread in the current channel"""
        await ctx.defer(ephemeral=True)
        await ctx.respond("Your private test thread")
        thread = await ctx.channel.create_thread(
            name=ctx.user.name + "'s private test conversation",
            auto_archive_duration=60,
        )
        await thread.send(
            f"<@{str(ctx.user.id)}> This is a private thread for testing. Only you and server admins can see this thread."
        )

    async def converse_command(
        self,
        ctx: discord.ApplicationContext,
        opener: str,
        opener_file: str,
        private: bool,
        minimal: bool,
        model: str,
        temperature: float,
        top_p: float,
        frequency_penalty: float,
        presence_penalty: float,
        use_threads: bool = True,  # Add this parameter
    ):
        """Command handler. Starts a conversation with the bot

        Args:
            ctx (discord.ApplicationContext): Command interaction
            opener (str): The first prompt to send in the conversation
            opener_file (str): A .txt or .json file which is appended before the opener
            private (bool): If the thread should be private
            minimal (bool): If a minimal starter should be used
            model (str): The openai model that should be used
            temperature (float): Sets the temperature override
            top_p (float): Sets the top p override
            frequency_penalty (float): Sets the frequency penalty override
            presence_penalty (float): Sets the presence penalty override
        """

        user = ctx.user

        # If we are in user input api keys mode, check if the user has entered their api key before letting them continue
        user_api_key = None
        if USER_INPUT_API_KEYS:
            user_api_key = await TextService.get_user_api_key(user.id, ctx, USER_KEY_DB)
            if not user_api_key:
                return

        if private:
            await ctx.defer(ephemeral=True)
        elif not private:
            await ctx.defer()

        # Check the opener for bad content.
        if PRE_MODERATE and opener is not None:
            if await Moderation.simple_moderate_and_respond(opener, ctx):
                return

        if use_threads:
            if private:
                embed_title = f"{user.name}'s private conversation with GPT"
                thread = await ctx.channel.create_thread(
                    name=embed_title,
                    auto_archive_duration=60,
                )
                target = thread
            else:
                embed_title = f"{user.name}'s conversation with GPT"
                message_embed = discord.Embed(title=embed_title, color=0x808080)
                message_thread = await ctx.send(embed=message_embed)
                thread = await message_thread.create_thread(
                    name=user.name + "'s conversation with GPT",
                    auto_archive_duration=60,
                )
                await ctx.respond("Conversation started.")
                target = thread
        else:
            # Check if this current channel is already in a conversation
            if ctx.channel.id in self.conversation_threads:
                await ctx.respond(
                    "There is already a conversation in this channel. Please finish that conversation before starting a new one."
                )
                return

            # Check if the user is permitted to start a conversation in full channels
            # check if any of the user role names match CHANNEL_CHAT_ROLES
            if CHANNEL_CHAT_ROLES and CHANNEL_CHAT_ROLES != [None]:
                if not any(
                    role.name.lower() in CHANNEL_CHAT_ROLES for role in ctx.user.roles
                ):
                    await ctx.respond(
                        "You are not permitted to start a conversation in this channel."
                    )
                    return

            target = ctx.channel
            if private:
                embed_title = f"{user.name}'s private conversation with GPT"
            else:
                embed_title = f"{user.name}'s conversation with GPT"

            embed = discord.Embed(title=embed_title, color=0x808080)
            await ctx.respond(embed=embed)

        self.conversation_threads[target.id] = Thread(target.id)
        self.conversation_threads[target.id].model = (
            self.model.model if not model else model
        )

        # Set the overrides for the conversation
        self.conversation_threads[target.id].set_overrides(
            temperature, top_p, frequency_penalty, presence_penalty
        )

        if opener or opener_file:
            user_id_normalized = ctx.author.id
        else:
            user_id_normalized = user.id
        if opener_file:
            if not opener_file.endswith((".txt", ".json")):
                opener_file = (
                    None  # Just start a regular thread if the file fails to load
                )
            else:
                # Load the file and read it into opener
                try:
                    opener_file = re.sub(
                        ".+(?=[\\//])", "", opener_file
                    )  # remove paths from the opener file
                    opener_file = EnvService.find_shared_file(
                        f"openers{separator}{opener_file}"
                    )
                    opener_file = await self.load_file(opener_file, ctx)
                    try:  # Try opening as json, if it fails it'll just pass the whole txt or json to the opener
                        opener_file = json.loads(opener_file)
                        temperature = opener_file.get("temperature", None)
                        top_p = opener_file.get("top_p", None)
                        frequency_penalty = opener_file.get("frequency_penalty", None)
                        presence_penalty = opener_file.get("presence_penalty", None)
                        self.conversation_threads[target.id].set_overrides(
                            temperature, top_p, frequency_penalty, presence_penalty
                        )
                        if (
                            not opener
                        ):  # if we only use opener_file then only pass on opener_file for the opening prompt
                            opener = opener_file.get("text", "error getting text")
                        else:
                            opener = (
                                opener_file.get("text", "error getting text") + opener
                            )
                    except Exception:  # Parse as just regular text
                        if not opener:
                            opener = opener_file
                        else:
                            opener = opener_file + opener
                except Exception:
                    opener_file = (
                        None  # Just start a regular thread if the file fails to load
                    )

        # Append the starter text for gpt to the user's history so it gets concatenated with the prompt later
        if minimal or opener_file or opener:
            self.conversation_threads[target.id].history.append(
                EmbeddedConversationItem(self.CONVERSATION_STARTER_TEXT_MINIMAL, 0)
            )
        elif not minimal:
            self.conversation_threads[target.id].history.append(
                EmbeddedConversationItem(self.CONVERSATION_STARTER_TEXT, 0)
            )

        # Set user as thread owner before sending anything that can error and leave the thread unowned
        self.conversation_thread_owners[user_id_normalized].append(target.id)
        overrides = self.conversation_threads[target.id].get_overrides()

        await target.send(f"<@{str(ctx.user.id)}> is the thread owner.")

        await target.send(
            embed=EmbedStatics.generate_conversation_embed(
                self.conversation_threads, target, opener, overrides
            )
        )

        # send opening
        if opener:
            self.conversation_threads[target.id].has_opener = True
            opener = await self.mention_to_username(ctx, opener)
            target_message = await target.send(
                embed=EmbedStatics.generate_opener_embed(opener)
            )
            if target.id in self.conversation_threads:
                self.awaiting_responses.append(user_id_normalized)
                if not self.pinecone_service:
                    self.conversation_threads[target.id].history.append(
                        EmbeddedConversationItem(
                            f"\n{ctx.author.display_name}: {opener} <|endofstatement|>\n",
                            0,
                        )
                    )
                self.awaiting_thread_responses.append(target.id)

                # ... (no other changes in the middle part of the function)

            overrides = Override(
                overrides["temperature"],
                overrides["top_p"],
                overrides["frequency_penalty"],
                overrides["presence_penalty"],
            )

            await TextService.encapsulated_send(
                self,
                target.id,
                opener
                if target.id not in self.conversation_threads or self.pinecone_service
                else "".join(
                    [item.text for item in self.conversation_threads[target.id].history]
                ),
                target_message,
                overrides=overrides,
                user=user,
                model=self.conversation_threads[target.id].model,
                custom_api_key=user_api_key,
            )
            self.awaiting_responses.remove(user_id_normalized)
            if target.id in self.awaiting_thread_responses:
                self.awaiting_thread_responses.remove(target.id)

    async def end_command(self, ctx: discord.ApplicationContext):
        """Command handler. Gets the user's thread and ends it"""
        await ctx.defer(ephemeral=True)
        user_id = ctx.user.id

        if ctx.channel.id in self.conversation_threads:
            try:
                await self.end_conversation(ctx)
            except Exception as e:
                print(e)
                traceback.print_exc()
        else:
            await ctx.respond(
                "This is not a conversation channel.", ephemeral=True, delete_after=10
            )

    async def setup_command(self, ctx: discord.ApplicationContext):
        """Command handler. Opens the setup modal"""
        if not USER_INPUT_API_KEYS:
            await ctx.respond(
                "This server doesn't support user input API keys.",
                ephemeral=True,
                delete_after=30,
            )
            return

        modal = SetupModal(user_key_db=USER_KEY_DB)
        await ctx.send_modal(modal)

    async def settings_command(
        self, ctx: discord.ApplicationContext, parameter: str = None, value: str = None
    ):
        """Command handler. Returns current settings or sets new values"""
        await ctx.defer()
        if parameter is None and value is None:
            await self.send_settings_text(ctx)
            return

        # If only one of the options are set, then this is invalid.
        if (
            parameter is None
            and value is not None
            or parameter is not None
            and value is None
        ):
            await ctx.respond(
                "Invalid settings command. Please use `/settings <parameter> <value>` to change a setting"
            )
            return

        # Otherwise, process the settings change
        await self.process_settings(ctx, parameter, value)

    async def settings_reset_command(self, ctx: discord.ApplicationContext):
        """Command handler. Resets all settings to default"""
        await ctx.defer()
        self.model.reset_settings()
        await ctx.respond("Settings reset to default")

    #
    # Text-based context menu commands from here
    #

    async def ask_gpt_action(self, ctx, message: discord.Message):
        """Message command. Return the message"""
        prompt = await self.mention_to_username(ctx, message.content)
        await self.ask_command(
            ctx,
            prompt=prompt,
            from_ask_action=prompt,
        )

    async def paraphrase_action(self, ctx, message: discord.Message):
        """Message command. paraphrase the current message content"""
        user = ctx.user
        prompt = await self.mention_to_username(ctx, message.content)
        from_other_action = prompt + "\nParaphrased:"

        # Construct the paraphrase prompt
        prompt = f"Paraphrase the following text. Maintain roughly the same text length after paraphrasing and the same tone of voice: {prompt} \nParaphrased:"

        tokens = self.model.usage_service.count_tokens(prompt)
        if tokens > self.model.max_tokens - 1000:
            await ctx.respond(
                f"This message is too long to paraphrase.",
                ephemeral=True,
                delete_after=10,
            )
            return

        await self.ask_command(
            ctx,
            prompt=prompt,
            from_other_action=from_other_action,
        )

    async def elaborate_action(self, ctx, message: discord.Message):
        """Message command. elaborate on the subject of the current message content"""
        user = ctx.user
        prompt = await self.mention_to_username(ctx, message.content)
        from_other_action = prompt + "\nElaboration:"

        # Construct the paraphrase prompt
        prompt = f"Elaborate with more information about the subject of the following message. Be objective and detailed and respond with elaborations only about the subject(s) of the message: {prompt} \n\nElaboration:"

        tokens = self.model.usage_service.count_tokens(prompt)
        if tokens > self.model.max_tokens - 500:
            await ctx.respond(
                f"This message is too long to elaborate on.",
                ephemeral=True,
                delete_after=10,
            )
            return

        await self.ask_command(
            ctx,
            prompt=prompt,
            from_other_action=from_other_action,
        )

    async def summarize_action(self, ctx, message: discord.Message):
        """Message command. elaborate on the subject of the current message content"""
        user = ctx.user
        prompt = await self.mention_to_username(ctx, message.content)
        from_other_action = (
            "Message at message link: " + message.jump_url + "\nSummarized:"
        )

        # Construct the paraphrase prompt
        prompt = f"Summarize the following message, be as short and concise as possible: {prompt} \n\nSummary:"

        tokens = self.model.usage_service.count_tokens(prompt)
        if tokens > self.model.max_tokens - 300:
            await ctx.respond(
                f"Your prompt is too long. It has {tokens} tokens, but the maximum is {self.model.max_tokens-300}.",
                ephemeral=True,
                delete_after=10,
            )
            return

        await self.ask_command(
            ctx,
            prompt=prompt,
            from_other_action=from_other_action,
        )


class ShareView(discord.ui.View):
    def __init__(
        self,
        converser_cog,
        conversation_id,
    ):
        super().__init__(timeout=3600)  # 1 hour interval to share the conversation.
        self.converser_cog = converser_cog
        self.conversation_id = conversation_id
        self.add_item(ShareButton(converser_cog, conversation_id))

    async def on_timeout(self):
        # Remove the button from the view/message
        self.clear_items()


class ShareButton(discord.ui.Button["ShareView"]):
    def __init__(self, converser_cog, conversation_id):
        super().__init__(
            style=discord.ButtonStyle.green,
            label="Share Conversation",
            custom_id="share_conversation",
        )
        self.converser_cog = converser_cog
        self.conversation_id = conversation_id

    async def callback(self, interaction: discord.Interaction):
        # Get the user
        try:
            id = await self.converser_cog.sharegpt_service.format_and_share(
                self.converser_cog.full_conversation_history[self.conversation_id],
                self.converser_cog.bot.user.default_avatar.url
                if not self.converser_cog.bot.user.avatar
                else self.converser_cog.bot.user.avatar.url,
            )
            url = f"https://shareg.pt/{id}"
            await interaction.response.send_message(
                embed=EmbedStatics.get_conversation_shared_embed(url)
            )
        except ValueError as e:
            traceback.print_exc()
            await interaction.response.send_message(
                embed=EmbedStatics.get_conversation_share_failed_embed(
                    "The ShareGPT API returned an error: " + str(e)
                ),
                ephemeral=True,
                delete_after=15,
            )
            return
        except Exception as e:
            traceback.print_exc()
            await interaction.response.send_message(
                embed=EmbedStatics.get_conversation_share_failed_embed(str(e)),
                ephemeral=True,
                delete_after=15,
            )
            return



import asyncio


class Message:
    def __init__(self, content, channel):
        self.content = content
        self.channel = channel

    # This function will be called by the bot to process the message queue
    @staticmethod
    async def process_message_queue(message_queue, PROCESS_WAIT_TIME, EMPTY_WAIT_TIME):
        while True:
            await asyncio.sleep(PROCESS_WAIT_TIME)
            # If the queue is empty, sleep for a short time before checking again
            if message_queue.empty():
                await asyncio.sleep(EMPTY_WAIT_TIME)
                continue

            # Get the next message from the queue
            message = await message_queue.get()

            # Send the message
            try:
                await message.channel.send(message.content)
            except Exception:
                pass

            # Sleep for a short time before processing the next message
            # This will prevent the bot from spamming messages too quickly
            await asyncio.sleep(PROCESS_WAIT_TIME)



import os

from services.environment_service import EnvService
import replicate


class BLIPModel:
    def __init__(self):
        # Try to get the replicate API key from the environment
        self.replicate_key = EnvService.get_replicate_api_key()
        # Set the environment REPLICATE_API_TOKEN to the replicate API key
        if self.replicate_key:
            os.environ["REPLICATE_API_TOKEN"] = self.replicate_key
            self.key_set = True
        else:
            self.key_set = False

    def get_is_usable(self):
        return self.key_set

    def ask_image_question(self, prompt, filepath):
        output = replicate.run(
            "andreasjansson/blip-2:4b32258c42e9efd4288bb9910bc532a69727f9acd26aa08e175713a0a857a608",
            input={"image": open(filepath, "rb"), "question": prompt},
        )
        return output



import datetime
import io
import os
import tempfile
import traceback
from typing import Optional, Dict, Any

import aiohttp
import re
import discord
from bs4 import BeautifulSoup
from discord.ext import pages
from langchain import (
    GoogleSearchAPIWrapper,
    WolframAlphaAPIWrapper,
    FAISS,
    InMemoryDocstore,
)
from langchain.agents import Tool, initialize_agent, AgentType
from langchain.chat_models import ChatOpenAI
from langchain.memory import ConversationBufferMemory, CombinedMemory
from langchain.requests import TextRequestsWrapper, Requests
from llama_index import (
    GPTSimpleVectorIndex,
    Document,
    SimpleDirectoryReader,
    ServiceContext,
    OpenAIEmbedding,
)
from llama_index.prompts.chat_prompts import CHAT_REFINE_PROMPT
from pydantic import Extra, BaseModel
from transformers import GPT2TokenizerFast

from models.embed_statics_model import EmbedStatics
from models.search_model import Search
from services.deletion_service import Deletion
from services.environment_service import EnvService
from services.moderations_service import Moderation
from services.text_service import TextService

from contextlib import redirect_stdout


async def capture_stdout(func, *args, **kwargs):
    buffer = io.StringIO()
    with redirect_stdout(buffer):
        result = await func(*args, **kwargs)
    captured_output = buffer.getvalue()
    buffer.close()
    return result, captured_output


ALLOWED_GUILDS = EnvService.get_allowed_guilds()
USER_INPUT_API_KEYS = EnvService.get_user_input_api_keys()
USER_KEY_DB = EnvService.get_api_db()
PRE_MODERATE = EnvService.get_premoderate()
GOOGLE_API_KEY = EnvService.get_google_search_api_key()
GOOGLE_SEARCH_ENGINE_ID = EnvService.get_google_search_engine_id()
OPENAI_API_KEY = EnvService.get_openai_token()
# Set the environment
os.environ["OPENAI_API_KEY"] = OPENAI_API_KEY
WOLFRAM_API_KEY = EnvService.get_wolfram_api_key()

vector_stores = {}


class RedoSearchUser:
    def __init__(self, ctx, query, search_scope, nodes, response_mode):
        self.ctx = ctx
        self.query = query
        self.search_scope = search_scope
        self.nodes = nodes
        self.response_mode = response_mode


class CustomTextRequestWrapper(BaseModel):
    """Lightweight wrapper around requests library.

    The main purpose of this wrapper is to always return a text output.
    """

    headers: Optional[Dict[str, str]] = None
    aiosession: Optional[aiohttp.ClientSession] = None
    tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def __init__(self, **data: Any):
        super().__init__(**data)

    @property
    def requests(self) -> Requests:
        return Requests(headers=self.headers, aiosession=self.aiosession)

    def get(self, url: str, **kwargs: Any) -> str:
        # the "url" field is actuall some input from the LLM, it is a comma separated string of the url and a boolean value and the original query
        url, use_gpt4, original_query = url.split(",")
        use_gpt4 = use_gpt4 == "True"
        """GET the URL and return the text."""
        text = self.requests.get(url, **kwargs).text

        # Load this text into BeautifulSoup, clean it up and only retain text content within <p> and <title> and <h1> type tags, get rid of all javascript and css too.
        soup = BeautifulSoup(text, "html.parser")

        # Decompose script, style, head, and meta tags
        for tag in soup(["script", "style", "head", "meta"]):
            tag.decompose()

        # Get remaining text from the soup object
        text = soup.get_text()

        # Clean up white spaces
        text = re.sub(r"\s+", " ", text)

        # If not using GPT-4 and the text token amount is over 3500, truncate it to 3500 tokens
        tokens = len(self.tokenizer(text)["input_ids"])
        print("The scraped text content is: " + text)
        if len(text) < 5:
            return "This website could not be scraped. I cannot answer this question."
        if (not use_gpt4 and tokens > 3000) or (use_gpt4 and tokens > 7000):
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
                f.write(text)
                f.close()
                document = SimpleDirectoryReader(input_files=[f.name]).load_data()
                embed_model = OpenAIEmbedding()
                service_context = ServiceContext.from_defaults(embed_model=embed_model)
                index = GPTSimpleVectorIndex.from_documents(
                    document, service_context=service_context, use_async=True
                )
                response_text = index.query(
                    original_query,
                    refine_template=CHAT_REFINE_PROMPT,
                    similarity_top_k=4,
                    response_mode="compact",
                )
                return response_text

        return text


class SearchService(discord.Cog, name="SearchService"):
    """Cog containing translation commands and retrieval of translation services"""

    def __init__(
        self,
        bot,
        gpt_model,
        usage_service,
        deletion_service,
        converser_cog,
    ):
        super().__init__()
        self.bot = bot
        self.usage_service = usage_service
        self.model = Search(gpt_model, usage_service)
        self.EMBED_CUTOFF = 2000
        self.redo_users = {}
        self.chat_agents = {}
        self.thread_awaiting_responses = []
        self.converser_cog = converser_cog
        # Make a mapping of all the country codes and their full country names:

    async def paginate_embed(
        self, response_text, user: discord.Member, original_link=None
    ):
        """Given a response text make embed pages and return a list of the pages."""

        response_text = [
            response_text[i : i + self.EMBED_CUTOFF]
            for i in range(0, len(response_text), self.EMBED_CUTOFF)
        ]
        pages = []
        first = False
        # Send each chunk as a message
        for count, chunk in enumerate(response_text, start=1):
            if not first:
                page = discord.Embed(
                    title="Search Results"
                    if not original_link
                    else "Follow-up results",
                    description=chunk,
                    url=original_link,
                )
                first = True
            else:
                page = discord.Embed(
                    title=f"Page {count}",
                    description=chunk,
                    url=original_link,
                )
            if user.avatar:
                page.set_footer(
                    text=f"Requested by {user.name}", icon_url=user.avatar.url
                )
            else:
                page.set_footer(
                    text=f"Requested by {user.name}", icon_url=user.default_avatar.url
                )
            pages.append(page)

        return pages

    async def paginate_chat_embed(self, response_text):
        """Given a response text make embed pages and return a list of the pages."""

        response_text = [
            response_text[i : i + 3500] for i in range(0, len(response_text), 7000)
        ]
        pages = []
        first = False
        # Send each chunk as a message
        for count, chunk in enumerate(response_text, start=1):
            if not first:
                page = discord.Embed(
                    title=f"{count}",
                    description=chunk,
                )
                first = True
            else:
                page = discord.Embed(
                    title=f"{count}",
                    description=chunk,
                )
            pages.append(page)

        return pages

    @discord.Cog.listener()
    async def on_message(self, message):
        # Check if the message is from a bot.
        if message.author.id == self.bot.user.id:
            return

        # Check if the message is from a guild.
        if not message.guild:
            return

        if message.content.strip().startswith("~"):
            return

        # if we are still awaiting a response from the agent, then we don't want to process the message.
        if message.channel.id in self.thread_awaiting_responses:
            resp_message = await message.reply(
                "Please wait for the agent to respond to a previous message first!"
            )
            deletion_time = datetime.datetime.now() + datetime.timedelta(seconds=5)
            deletion_time = deletion_time.timestamp()

            original_deletion_message = Deletion(message, deletion_time)
            deletion_message = Deletion(resp_message, deletion_time)
            await self.converser_cog.deletion_queue.put(deletion_message)
            await self.converser_cog.deletion_queue.put(original_deletion_message)
            return

        # Pre moderation
        if PRE_MODERATE:
            if await Moderation.simple_moderate_and_respond(message.content, message):
                await message.delete()
                return

        prompt = message.content.strip()

        # If the message channel is in self.chat_agents, then we delegate the message to the agent.
        if message.channel.id in self.chat_agents:
            if prompt in ["stop", "end", "quit", "exit"]:
                await message.reply("Ending chat session.")
                self.chat_agents.pop(message.channel.id)

                # close the thread
                thread = await self.bot.fetch_channel(message.channel.id)
                await thread.edit(name="Closed-GPT")
                await thread.edit(archived=True)
                return

            self.thread_awaiting_responses.append(message.channel.id)

            try:
                await message.channel.trigger_typing()
            except:
                pass

            agent = self.chat_agents[message.channel.id]
            used_tools = []
            try:
                # Start listening to STDOUT before this call. We wanna track all the output for this specific call below
                response, stdout_output = await capture_stdout(
                    self.bot.loop.run_in_executor, None, agent.run, prompt
                )
                response = str(response)

                try:
                    print(stdout_output)
                except:
                    traceback.print_exc()
                    stdout_output = ""

                if "Wolfram-Tool" in stdout_output:
                    used_tools.append("Wolfram Alpha")
                if "Search-Tool" in stdout_output:
                    used_tools.append("Google Search")
                if "Web-Crawling-Tool" in stdout_output:
                    used_tools.append("Web Crawler")

            except Exception as e:
                # Try again one more time
                try:
                    response = await self.bot.loop.run_in_executor(
                        None, agent.run, prompt
                    )
                except Exception as e:
                    response = f"Error: {e}"
                    traceback.print_exc()

            if len(response) > 2000:
                embed_pages = await self.paginate_chat_embed(response)
                paginator = pages.Paginator(
                    pages=embed_pages,
                    timeout=None,
                    author_check=False,
                )
                await paginator.respond(message)
            else:
                response = response.replace("\\n", "\n")
                # Build a response embed
                response_embed = discord.Embed(
                    title="",
                    description=response,
                    color=0x808080,
                )
                if len(used_tools) > 0:
                    response_embed.set_footer(
                        text="Used tools: " + ", ".join(used_tools)
                    )
                await message.reply(embed=response_embed)

            self.thread_awaiting_responses.remove(message.channel.id)

    async def search_chat_command(
        self, ctx: discord.ApplicationContext, search_scope=2, use_gpt4: bool = False
    ):
        embed_title = f"{ctx.user.name}'s internet-connected conversation with GPT"
        message_embed = discord.Embed(
            title=embed_title,
            description=f"The agent will visit and browse **{search_scope}** link(s) every time it needs to access the internet.\nCrawling is enabled, send the bot a link for it to access it!\nModel: {'gpt-3.5-turbo' if not use_gpt4 else 'GPT-4'}\n\nType `end` to stop the conversation",
            color=0x808080,
        )
        message_thread = await ctx.send(embed=message_embed)
        thread = await message_thread.create_thread(
            name=ctx.user.name + "'s internet-connected conversation with GPT",
            auto_archive_duration=60,
        )
        await ctx.respond("Conversation started.")
        print("The search scope is " + str(search_scope) + ".")

        # Make a new agent for this user to chat.
        search = GoogleSearchAPIWrapper(
            google_api_key=GOOGLE_API_KEY,
            google_cse_id=GOOGLE_SEARCH_ENGINE_ID,
            k=search_scope,
        )

        requests = CustomTextRequestWrapper()

        tools = [
            Tool(
                name="Search-Tool",
                func=search.run,
                description="useful when you need to answer questions about current events or retrieve information about a topic that may require the internet.",
            ),
            # The requests tool
            Tool(
                name="Web-Crawling-Tool",
                func=requests.get,
                description=f"Useful for when the user provides you with a website link, use this tool to crawl the website and retrieve information from it. The input to this tool is a comma separated list of three values, the first value is the link to crawl for, and the second value is the value of use_gpt4, which is {use_gpt4}, and the third value is the original question that the user asked. For example, an input could be 'https://google.com', False, 'What is this webpage?'",
            ),
        ]

        # Try to add wolfram tool
        try:
            wolfram = WolframAlphaAPIWrapper(wolfram_alpha_appid=WOLFRAM_API_KEY)
            tools.append(
                Tool(
                    name="Wolfram-Tool",
                    func=wolfram.run,
                    description="useful when you need to answer questions about math, solve equations, do proofs, mathematical science questions, science questions, and when asked to do numerical based reasoning.",
                )
            )
            print("Wolfram tool added to internet-connected conversation agent.")
        except Exception:
            traceback.print_exc()
            print("Wolfram tool not added to internet-connected conversation agent.")

        memory = ConversationBufferMemory(
            memory_key="chat_history", return_messages=True
        )

        if use_gpt4:
            llm = ChatOpenAI(
                model="gpt-4", temperature=0.7, openai_api_key=OPENAI_API_KEY
            )
        else:
            llm = ChatOpenAI(
                model="gpt-3.5-turbo", temperature=0.7, openai_api_key=OPENAI_API_KEY
            )

        agent_chain = initialize_agent(
            tools,
            llm,
            agent=AgentType.CHAT_CONVERSATIONAL_REACT_DESCRIPTION,
            verbose=True,
            memory=memory,
            max_execution_time=120,
            max_iterations=4,
            early_stopping_method="generate",
        )

        self.chat_agents[thread.id] = agent_chain

    async def search_command(
        self,
        ctx: discord.ApplicationContext,
        query,
        search_scope,
        nodes,
        deep,
        response_mode,
        model="gpt-3.5-turbo",
        multistep=False,
        redo=None,
        from_followup=None,
        followup_user=None,
    ):
        """Command handler for the search command"""
        await ctx.defer() if not redo else None

        # Check the opener for bad content.
        if PRE_MODERATE:
            if await Moderation.simple_moderate_and_respond(query, ctx):
                return

        user_api_key = None
        if USER_INPUT_API_KEYS:
            user_api_key = await TextService.get_user_api_key(
                ctx.user.id, ctx, USER_KEY_DB
            )
            if not user_api_key:
                return

        if (
            not EnvService.get_google_search_api_key()
            or not EnvService.get_google_search_engine_id()
        ):
            await ctx.respond(
                embed=EmbedStatics.get_search_failure_embed(
                    str("The search service is not enabled on this server.")
                ),
            )
            return

        try:
            response, refined_text = await self.model.search(
                ctx,
                query,
                user_api_key,
                search_scope,
                nodes,
                deep,
                response_mode,
                model,
                multistep,
            )
        except ValueError as e:
            traceback.print_exc()
            await ctx.respond(
                embed=EmbedStatics.get_search_failure_embed(str(e)),
                ephemeral=True,
            )
            return
        except Exception as e:
            await ctx.respond(
                embed=EmbedStatics.get_search_failure_embed(str(e)), ephemeral=True
            )
            traceback.print_exc()
            return

        url_extract_pattern = "https?:\\/\\/(?:www\\.)?[-a-zA-Z0-9@:%._\\+~#=]{1,256}\\.[a-zA-Z0-9()]{1,6}\\b(?:[-a-zA-Z0-9()@:%_\\+.~#?&\\/=]*)"
        urls = re.findall(
            url_extract_pattern,
            str(response.get_formatted_sources(length=200)),
            flags=re.IGNORECASE,
        )
        urls = "\n".join(f"<{url}>" for url in urls)

        # Deduplicate the urls
        urls = "\n".join(dict.fromkeys(urls.split("\n")))

        if from_followup:
            original_link, followup_question = (
                from_followup.original_link,
                from_followup.followup_question,
            )
            query_response_message = f"**Question:**\n\n`{followup_question}`\n\n**Google Search Query**\n\n`{refined_text.strip()}`\n\n**Final Answer:**\n\n{response.response.strip()}\n\n**Sources:**\n{urls}"
        else:
            query_response_message = f"**Question:**\n\n`{query.strip()}`\n\n**Google Search Query**\n\n`{refined_text.strip()}`\n\n**Final Answer:**\n\n{response.response.strip()}\n\n**Sources:**\n{urls}"
        query_response_message = query_response_message.replace(
            "<|endofstatement|>", ""
        )
        query_response_message = query_response_message.replace(
            "Answer to original:\n", ""
        )
        query_response_message = query_response_message.replace(
            "Answer to follow-up:\n", ""
        )

        # If the response is too long, lets paginate using the discord pagination
        # helper
        embed_pages = await self.paginate_embed(
            query_response_message,
            ctx.user if not followup_user else followup_user,
            original_link if from_followup else None,
        )
        paginator = pages.Paginator(
            pages=embed_pages,
            timeout=None,
            author_check=False,
            custom_view=SearchView(ctx, self, query_response_message),
        )

        self.redo_users[ctx.user.id] = RedoSearchUser(
            ctx, query, search_scope, nodes, response_mode
        )

        await paginator.respond(ctx.interaction)


class SearchView(discord.ui.View):
    def __init__(
        self,
        ctx,
        search_cog,
        response_text,
    ):
        super().__init__(timeout=None)  # No timeout
        self.search_cog = search_cog
        self.ctx = ctx
        self.response_text = response_text
        self.add_item(RedoButton(self.ctx, self.search_cog))
        self.add_item(FollowupButton(self.ctx, self.search_cog, self.response_text))


# A view for a follow-up button
class FollowupButton(discord.ui.Button["SearchView"]):
    def __init__(self, ctx, search_cog, response_text):
        super().__init__(label="Follow Up", style=discord.ButtonStyle.green)
        self.search_cog = search_cog
        self.ctx = ctx
        self.response_text = response_text

    async def callback(self, interaction: discord.Interaction):
        """Send the followup modal"""
        await interaction.response.send_modal(
            modal=FollowupModal(self.ctx, self.search_cog, self.response_text)
        )


# A view for a redo button
class RedoButton(discord.ui.Button["SearchView"]):
    def __init__(self, ctx, search_cog):
        super().__init__(
            style=discord.ButtonStyle.danger,
            label="Redo",
            custom_id="redo_search_button",
        )
        self.ctx = ctx
        self.search_cog = search_cog

    async def callback(self, interaction: discord.Interaction):
        """Redo the search"""
        await interaction.response.send_message(
            embed=EmbedStatics.get_search_redo_progress_embed(),
            ephemeral=True,
            delete_after=15,
        )
        await self.search_cog.search_command(
            self.search_cog.redo_users[self.ctx.user.id].ctx,
            self.search_cog.redo_users[self.ctx.user.id].query,
            self.search_cog.redo_users[self.ctx.user.id].search_scope,
            self.search_cog.redo_users[self.ctx.user.id].nodes,
            deep=False,
            redo=True,
            response_mode=self.search_cog.redo_users[self.ctx.user.id].response_mode,
        )


class FollowupData:
    def __init__(self, original_link, followup_question):
        self.original_link = original_link
        self.followup_question = followup_question


# The modal for following up
class FollowupModal(discord.ui.Modal):
    def __init__(self, ctx, search_cog, response_text) -> None:
        super().__init__(title="Search Follow-up")
        # Get the argument named "user_key_db" and save it as USER_KEY_DB
        self.search_cog = search_cog
        self.ctx = ctx
        self.response_text = response_text

        self.add_item(
            discord.ui.InputText(
                label="What other questions do you have?",
                placeholder="",
            )
        )

    async def callback(self, interaction: discord.Interaction):
        await interaction.response.defer()
        query = self.search_cog.redo_users[self.ctx.user.id].query

        # In the response text, get only the text between "**Final Answer:**" and "**Sources:**"
        self.response_text = self.response_text.split("**Final Answer:**")[1].split(
            "**Sources:**"
        )[0]

        # Build the context
        context_text = (
            "Original question: "
            + query
            + "\n"
            + "Answer to original: "
            + self.response_text
            + "\n"
            + "Follow-up question: "
            + self.children[0].value
        )

        # Get the link of the message that the user interacted on
        message_link = f"https://discord.com/channels/{interaction.guild_id}/{interaction.channel_id}/{interaction.message.id}"

        await self.search_cog.search_command(
            self.search_cog.redo_users[self.ctx.user.id].ctx,
            context_text,
            self.search_cog.redo_users[self.ctx.user.id].search_scope,
            self.search_cog.redo_users[self.ctx.user.id].nodes,
            deep=False,
            redo=True,
            from_followup=FollowupData(message_link, self.children[0].value),
            response_mode=self.search_cog.redo_users[self.ctx.user.id].response_mode,
            followup_user=interaction.user,
        )



import os
import sys
import traceback
from pathlib import Path
from typing import Union

from dotenv import load_dotenv
from sqlitedict import SqliteDict


def app_root_path():
    app_path = Path(sys.argv[0]).resolve()
    try:
        if app_path.parent.name == "bin":  # Installed in unixy hierachy
            return app_path.parents[1]
    except IndexError:
        pass
    return app_path.parent


# None will let direnv do its' thing
env_paths = [Path(".env"), app_root_path() / "etc/environment", None]

for env_path in env_paths:
    print("Loading environment from " + str(env_path))
    load_dotenv(dotenv_path=env_path)


class EnvService:
    # To be expanded upon later!
    def __init__(self):
        self.env = {}

    @staticmethod
    def environment_path_with_fallback(env_name, relative_fallback=None):
        directory = os.getenv(env_name)
        if directory is not None:
            return Path(directory).resolve()

        if relative_fallback:
            app_relative = (app_root_path() / relative_fallback).resolve()
            if app_relative.exists():
                return app_relative

        return Path.cwd()

    @staticmethod
    def save_path():
        share_dir = os.getenv("SHARE_DIR")
        if share_dir is not None:
            return Path(share_dir)
        return app_root_path()

    @staticmethod
    def find_shared_file(file_name):
        share_file_paths = []
        share_dir = os.getenv("SHARE_DIR")
        if share_dir is not None:
            share_file_paths.append(Path(share_dir) / file_name)

        share_file_paths.extend(
            [
                app_root_path() / "share" / file_name,
                app_root_path() / file_name,
                Path(file_name),
            ]
        )

        for share_file_path in share_file_paths:
            if share_file_path.exists():
                return share_file_path.resolve()

        raise ValueError(f"Unable to find shared data file {file_name}")

    @staticmethod
    def get_allowed_guilds():
        # ALLOWED_GUILDS is a comma separated list of guild ids
        # It can also just be one guild ID
        # Read these allowed guilds and return as a list of ints
        try:
            allowed_guilds = os.getenv("ALLOWED_GUILDS")
        except Exception:
            allowed_guilds = None

        if allowed_guilds is None:
            raise ValueError(
                "ALLOWED_GUILDS is not defined properly in the environment file!"
                "Please copy your server's guild ID and put it into ALLOWED_GUILDS in the .env file."
                'For example a line should look like: `ALLOWED_GUILDS="971268468148166697"`'
            )

        allowed_guilds = (
            allowed_guilds.split(",") if "," in allowed_guilds else [allowed_guilds]
        )
        allowed_guilds = [int(guild) for guild in allowed_guilds]
        return allowed_guilds

    @staticmethod
    def get_admin_roles():
        # ADMIN_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            admin_roles = os.getenv("ADMIN_ROLES")
        except Exception:
            admin_roles = None

        if admin_roles is None:
            print(
                "ADMIN_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into ADMIN_ROLES in the .env file."
                'For example a line should look like: `ADMIN_ROLES="Admin"`'
            )
            print("Defaulting to allowing all users to use admin commands...")
            return [None]

        admin_roles = (
            admin_roles.lower().split(",")
            if "," in admin_roles
            else [admin_roles.lower()]
        )
        return admin_roles

    @staticmethod
    def get_dalle_roles():
        # DALLE_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            dalle_roles = os.getenv("DALLE_ROLES")
        except Exception:
            dalle_roles = None

        if dalle_roles is None:
            print(
                "DALLE_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into DALLE_ROLES in the .env file."
                'For example a line should look like: `DALLE_ROLES="Dalle"`'
            )
            print("Defaulting to allowing all users to use Dalle commands...")
            return [None]

        dalle_roles = (
            dalle_roles.lower().split(",")
            if "," in dalle_roles
            else [dalle_roles.lower()]
        )
        return dalle_roles

    @staticmethod
    def get_translator_roles():
        # TRANSLATOR_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            translator_roles = os.getenv("TRANSLATOR_ROLES")
        except Exception:
            translator_roles = None

        if translator_roles is None:
            print(
                "TRANSLATOR_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into TRANSLATOR in the .env file."
                'For example a line should look like: `TRANSLATOR_ROLES="Translate"`'
            )
            print("Defaulting to allowing all users to use Translator commands...")
            return [None]

        translator_roles = (
            translator_roles.lower().split(",")
            if "," in translator_roles
            else [translator_roles.lower()]
        )
        return translator_roles

    @staticmethod
    def get_search_roles():
        # SEARCH_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            search_roles = os.getenv("SEARCH_ROLES")
        except Exception:
            search_roles = None

        if search_roles is None:
            print(
                "SEARCH_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into SEARCH in the .env file."
                'For example a line should look like: `SEARCH_ROLES="Translate"`'
            )
            print("Defaulting to allowing all users to use Search commands...")
            return [None]

        search_roles = (
            search_roles.lower().split(",")
            if "," in search_roles
            else [search_roles.lower()]
        )
        return search_roles

    @staticmethod
    def get_gpt_roles():
        # GPT_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            gpt_roles = os.getenv("GPT_ROLES")
        except Exception:
            gpt_roles = None

        if gpt_roles is None:
            print(
                "GPT_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into GPT_ROLES in the .env file."
                'For example a line should look like: `GPT_ROLES="Gpt"`'
            )
            print("Defaulting to allowing all users to use GPT commands...")
            return [None]

        gpt_roles = (
            gpt_roles.lower().strip().split(",")
            if "," in gpt_roles
            else [gpt_roles.lower()]
        )
        return gpt_roles

    @staticmethod
    def get_index_roles():
        # INDEX_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            index_roles = os.getenv("INDEX_ROLES")
        except Exception:
            traceback.print_exc()
            index_roles = None

        if index_roles is None:
            print(
                "INDEX_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into INDEX_ROLES in the .env file."
                'For example a line should look like: `INDEX_ROLES="Gpt"`'
            )
            print("Defaulting to allowing all users to use Index commands...")
            return [None]

        index_roles = (
            index_roles.lower().strip().split(",")
            if "," in index_roles
            else [index_roles.lower()]
        )
        return index_roles

    @staticmethod
    def get_channel_chat_roles():
        # CHANNEL_CHAT_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            cc_roles = os.getenv("CHANNEL_CHAT_ROLES")
        except Exception:
            cc_roles = None

        if cc_roles is None:
            print(
                "CHANNEL_CHAT_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into CHANNEL_CHAT_ROLES in the .env file."
                'For example a line should look like: `CHANNEL_CHAT_ROLES="Gpt"`'
            )
            print(
                "Defaulting to allowing all users to make conversations in full channels..."
            )
            return [None]

        cc_roles = (
            cc_roles.lower().strip().split(",")
            if "," in cc_roles
            else [cc_roles.lower()]
        )
        return cc_roles

    @staticmethod
    def get_channel_instruction_roles():
        # CHANNEL_INSTRUCTION_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            cc_roles = os.getenv("CHANNEL_INSTRUCTION_ROLES")
        except Exception:
            cc_roles = None

        if cc_roles is None:
            print(
                "CHANNEL_INSTRUCTION_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into CHANNEL_INSTRUCTION_ROLES in the .env file."
                'For example a line should look like: `CHANNEL_INSTRUCTION_ROLES="Gpt"`'
            )
            print(
                "Defaulting to allowing all users to set instructions for channels..."
            )
            return [None]

        cc_roles = (
            cc_roles.lower().strip().split(",")
            if "," in cc_roles
            else [cc_roles.lower()]
        )
        return cc_roles

    @staticmethod
    def get_welcome_message():
        # WELCOME_MESSAGE is a default string used to welcome new members to the server if GPT is not available.
        # The string can be blank but this is not advised. If a string cannot be found in the .env file, the below string is used.
        # The string is DMd to the new server member as part of an embed.
        try:
            welcome_message = os.getenv("WELCOME_MESSAGE")
        except Exception:
            welcome_message = "Hi there! Welcome to our Discord server!"
        return welcome_message

    @staticmethod
    def get_moderations_alert_channel():
        # MODERATIONS_ALERT_CHANNEL is a channel id where moderation alerts are sent to
        # The string can be blank but this is not advised. If a string cannot be found in the .env file, the below string is used.
        try:
            moderations_alert_channel = os.getenv("MODERATIONS_ALERT_CHANNEL")
        except Exception:
            moderations_alert_channel = None
        return moderations_alert_channel

    @staticmethod
    def get_user_input_api_keys():
        try:
            user_input_api_keys = os.getenv("USER_INPUT_API_KEYS")
            if user_input_api_keys.lower().strip() == "true":
                return True
            return False
        except Exception:
            return False

    @staticmethod
    def get_premoderate():
        try:
            pre_moderate = os.getenv("PRE_MODERATE")
            if pre_moderate.lower().strip() == "true":
                return True
            return False
        except Exception:
            return False

    @staticmethod
    def get_force_english():
        try:
            force_english = os.getenv("FORCE_ENGLISH")
            if force_english.lower().strip() == "true":
                return True
            return False
        except Exception:
            return False

    @staticmethod
    def get_custom_bot_name():
        try:
            custom_bot_name = os.getenv("CUSTOM_BOT_NAME") + ": "
            return custom_bot_name
        except Exception:
            return "GPTie: "

    @staticmethod
    def get_health_service_enabled():
        try:
            user_input_api_keys = os.getenv("HEALTH_SERVICE_ENABLED")
            if user_input_api_keys.lower().strip() == "true":
                return True
            return False
        except Exception:
            return False

    @staticmethod
    def get_bot_is_taggable():
        try:
            user_input_api_keys = os.getenv("BOT_TAGGABLE")
            if user_input_api_keys.lower().strip() == "true":
                return True
            return False
        except Exception:
            return False

    @staticmethod
    def get_user_key_db_path() -> Union[Path, None]:
        try:
            user_key_db_path = os.getenv("USER_KEY_DB_PATH")
            if user_key_db_path is None:
                return None
            return Path(user_key_db_path)
        except Exception:
            return None

    @staticmethod
    def get_api_db():
        user_input_api_keys = EnvService.get_user_input_api_keys()
        user_key_db = None
        if user_input_api_keys:
            print(
                "This server was configured to enforce user input API keys. Doing the required database setup now"
            )
            # Get user_key_db from environment variable
            user_key_db_path = EnvService.get_user_key_db_path()
            # Check if user_key_db_path is valid
            if not user_key_db_path:
                print(
                    "No user key database path was provided. Defaulting to user_key_db.sqlite"
                )
                user_key_db_path = EnvService.find_shared_file("user_key_db.sqlite")
            else:
                # append "user_key_db.sqlite" to USER_KEY_DB_PATH if it doesn't already end with .sqlite
                if not user_key_db_path.match("*.sqlite"):
                    # append "user_key_db.sqlite" to USER_KEY_DB_PATH
                    user_key_db_path = user_key_db_path / "user_key_db.sqlite"
            user_key_db = SqliteDict(user_key_db_path)
            print("Retrieved/created the user key database")
            return user_key_db
        return user_key_db

    @staticmethod
    def get_bypass_roles():
        # GPT_ROLES is a comma separated list of string roles
        # It can also just be one role
        # Read these allowed roles and return as a list of strings
        try:
            bypass_roles = os.getenv("CHAT_BYPASS_ROLES")
        except Exception:
            bypass_roles = None

        if bypass_roles is None:
            print(
                "CHAT_BYPASS_ROLES is not defined properly in the environment file!"
                "Please copy your server's role and put it into CHAT_BYPASS_ROLES in the .env file."
                'For example a line should look like: `CHAT_BYPASS_ROLES="bypass"`'
            )
            print("Defaulting to allowing NO ONE to bypass chat moderation")
            return [None]

        bypass_roles = (
            bypass_roles.lower().strip().split(",")
            if "," in bypass_roles
            else [bypass_roles.lower()]
        )
        return bypass_roles

    @staticmethod
    def get_deepl_token():
        try:
            deepl_token = os.getenv("DEEPL_TOKEN")
            return deepl_token
        except Exception:
            return None

    @staticmethod
    def get_openai_token():
        try:
            openai_token = os.getenv("OPENAI_TOKEN")
            return openai_token
        except Exception:
            raise ValueError(
                "OPENAI_TOKEN is not defined properly in the environment file! The bot cannot start without this token."
            )

    @staticmethod
    def get_wolfram_api_key():
        try:
            openai_token = os.getenv("WOLFRAM_API_KEY")
            return openai_token
        except Exception:
            print(
                "WOLFRAM_API_KEY is not defined properly in the environment file! The bot cannot use /internet chat's wolfram functionalities without this"
            )
            return None

    @staticmethod
    def get_openai_organization():
        try:
            openai_org = os.getenv("OPENAI_ORGANIZATION")
            return openai_org
        except Exception:
            return None

    @staticmethod
    def get_google_search_api_key():
        try:
            google_search_api_key = os.getenv("GOOGLE_SEARCH_API_KEY")
            return google_search_api_key
        except Exception:
            return None

    @staticmethod
    def get_replicate_api_key():
        try:
            replicate_key = os.getenv("REPLICATE_API_KEY")
            return replicate_key
        except Exception:
            return None

    @staticmethod
    def get_google_search_engine_id():
        try:
            google_search_engine_id = os.getenv("GOOGLE_SEARCH_ENGINE_ID")
            return google_search_engine_id
        except Exception:
            return None

    @staticmethod
    def get_pinecone_region():
        try:
            pinecone_region = os.getenv("PINECONE_REGION")
            return pinecone_region
        except Exception:
            return "us-west1-gcp"

    @staticmethod
    def get_max_search_price():
        try:
            search_price = float(os.getenv("MAX_SEARCH_PRICE"))
            return search_price
        except Exception:
            return 1.00

    @staticmethod
    def get_max_deep_compose_price():
        try:
            deep_compose_price = float(os.getenv("MAX_DEEP_COMPOSE_PRICE"))
            return deep_compose_price
        except Exception:
            return 3.00
